#!/usr/bin/perl
#remote_blast_client.pl
#Version 6.1
#
#Written by Paul Stothard, University of Alberta.
#stothard@ualberta.ca

use warnings;
use strict;
use Getopt::Long;
Getopt::Long::Configure('no_ignore_case');
use LWP::UserAgent;
use LWP::Protocol::https;
use HTTP::Request::Common;
use XML::DOM;
use Data::Dumper;

my %settings = (
    PROGRAM      => undef,
    DATABASE     => undef,
    EXPECT       => 10,
    WORD_SIZE    => undef,
    HITLIST_SIZE => 5,
    FILTER       => "F",
    OUTPUTFILE   => undef,
    INPUTFILE    => undef,
    INPUTTYPE    => undef,
    ENTREZ_DB    => undef,
    ENTREZ_QUERY => undef,
    ALIGN_TYPE   => undef,
    BLAST_URL    => "https://blast.ncbi.nlm.nih.gov/Blast.cgi",
    ENTREZ_URL =>
        "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?",
    HSP_MAX                      => undef,
    ERROR_RETRY                  => 5,
    MIN_HSP_LENGTH               => undef,
    MIN_HSP_PROP                 => undef,
    MIN_SCORE                    => undef,
    MIN_IDENTITY                 => undef,
    FETCH_DESC                   => 'T',
    QUERY_GENETIC_CODE           => 1,
    DATABASE_GENETIC_CODE        => 1,
    BROWSER                      => undef,
    MAX_BYTES_RESPONSE           => 5000000,
    COMPOSITION_BASED_STATISTICS => undef,
    ACCESSION                    => 'F',
    XML                          => 'F',
    HSP_LABEL                    => 'F',
    HSPS_PER_HIT                 => undef,
    MEGABLAST                    => 'false'
);

my $blastType = undef;
my $wordSize  = undef;

GetOptions(
    'i|input_file=s'            => \$settings{INPUTFILE},
    'o|output_file=s'           => \$settings{OUTPUTFILE},
    'b|blast_program=s'         => \$blastType,
    'e|entrez_query=s'          => \$settings{ENTREZ_QUERY},
    'd|database=s'              => \$settings{DATABASE},
    'h|hsps=i'                  => \$settings{HSP_MAX},
    'l|filter=s'                => \$settings{FILTER},
    'a|min_hsp_length=i'        => \$settings{MIN_HSP_LENGTH},
    'p|min_hsp_prop=f'          => \$settings{MIN_HSP_PROP},
    's|min_score=i'             => \$settings{MIN_SCORE},
    'n|min_identity=f'          => \$settings{MIN_IDENTITY},
    'x|expect=f'                => \$settings{EXPECT},
    't|hit_list_size=i'         => \$settings{HITLIST_SIZE},
    'f|fetch_description=s'     => \$settings{FETCH_DESC},
    'Q|query_genetic_code=i'    => \$settings{QUERY_GENETIC_CODE},
    'D|database_genetic_code=i' => \$settings{DATABASE_GENETIC_CODE},
    'accession=s'               => \$settings{ACCESSION},
    'xml=s'                     => \$settings{XML},
    'hsp_label=s'               => \$settings{HSP_LABEL},
    'hsps_per_hit=i'            => \$settings{HSPS_PER_HIT},
    'W|word_size=i'             => \$wordSize
);

if ( !( defined($blastType) ) ) {
    print "------------------------------------------------------------\n";
    print "Please enter a number to indicated the type of BLAST search\n";
    print "you want to perform:.\n";
    print "1 - Nucleotide-nucleotide BLAST (blastn).\n";
    print "2 - Protein-protein BLAST (blastp).\n";
    print "3 - Translated query vs protein database (blastx).\n";
    print "4 - Protein query vs translated database (tblastn).\n";
    print "5 - Translated query vs. translated database (tblastx).\n";
    print "6 - Nucleotide-nucleotide BLAST (megablast).\n";
    print "------------------------------------------------------------\n";
    $blastType = <STDIN>;
    chomp($blastType);

    if ( $blastType =~ m/(\d)/ ) {
        $blastType = $1;
    }
    else {
        die("Please enter a digit between 1 and 5.\n");
    }
    if ( ( $blastType < 1 ) || ( $blastType > 5 ) ) {
        die("Please enter a digit between 1 and 5.\n");
    }
}

if ( !( defined( $settings{DATABASE} ) ) ) {
    print "------------------------------------------------------------\n";
    print "Please enter the name of the database you wish to search.\n";
    print "------------------------------------------------------------\n";
    $settings{DATABASE} = <STDIN>;
    chomp( $settings{DATABASE} );
}

if (   ( !( defined( $settings{INPUTFILE} ) ) )
    && ( !( defined( $settings{ENTREZ_QUERY} ) ) ) )
{
    print "------------------------------------------------------------\n";
    print "Enter an entrez query or press enter to search all sequences.\n";
    print "For example, the term 'bacteria[Organism]' restricts the\n";
    print "search to bacteria sequences only.\n";
    print "------------------------------------------------------------\n";
    $settings{ENTREZ_QUERY} = <STDIN>;
    chomp( $settings{ENTREZ_QUERY} );
}

_setDefaults( $blastType, \%settings );

if ( !( defined( $settings{INPUTFILE} ) ) ) {
    print "------------------------------------------------------------\n";
    print "Enter the name of the FASTA format "
        . $settings{INPUTTYPE}
        . " sequence file\n";
    print "that contains your query sequences.\n";
    print "------------------------------------------------------------\n";
    $settings{INPUTFILE} = <STDIN>;
    chomp( $settings{INPUTFILE} );
}

my $inputLessExtentions = $settings{INPUTFILE};
if ( $settings{INPUTFILE} =~ m/(^[^\.]+)/g ) {
    $inputLessExtentions = $1;
}

if ( !( defined( $settings{OUTPUTFILE} ) ) ) {
    $settings{OUTPUTFILE} = $inputLessExtentions . "_" . "results.tab";

    print "------------------------------------------------------------\n";
    print "The results of this "
        . $settings{PROGRAM}
        . " search will be written to\n";
    print "a file called " . $settings{OUTPUTFILE} . ".\n";
    print "Start the search? (y or n) y\n";
    print "------------------------------------------------------------\n";
    my $continue = <STDIN>;
    chomp($continue);

    if ( $continue =~ m/n/i ) {
        exit(0);
    }
}

$settings{HITLIST_SIZE}       = _get_integer( $settings{HITLIST_SIZE} );
$settings{HSP_MAX}            = _get_integer( $settings{HSP_MAX} );
$settings{MIN_HSP_LENGTH}     = _get_integer( $settings{MIN_HSP_LENGTH} );
$settings{MIN_HSP_PROP}       = _get_real( $settings{MIN_HSP_PROP} );
$settings{MIN_SCORE}          = _get_integer( $settings{MIN_SCORE} );
$settings{MIN_IDENTITY}       = _get_real( $settings{MIN_IDENTITY} );
$settings{EXPECT}             = _get_real( $settings{EXPECT} );
$settings{HSPS_PER_HIT}       = _get_integer( $settings{HSPS_PER_HIT} );
$settings{QUERY_GENETIC_CODE} = _get_integer( $settings{QUERY_GENETIC_CODE} );
$settings{DATABASE_GENETIC_CODE}
    = _get_integer( $settings{DATABASE_GENETIC_CODE} );
$wordSize = _get_integer($wordSize);

if ( defined($wordSize) ) {
    $settings{WORD_SIZE} = $wordSize;
}

$settings{BROWSER} = LWP::UserAgent->new(
    ssl_opts => { verify_hostname => 0 },
    protocols_allowed => ['https'],
);
$settings{BROWSER}->timeout(30);
$settings{BROWSER}->max_size( $settings{MAX_BYTES_RESPONSE} );

open( OUTFILE, ">" . $settings{OUTPUTFILE} ) or die("Cannot open file : $!");
print( OUTFILE
        "#-------------------------------------------------------------------------------------------------------------------------------------------------\n"
);
print(    OUTFILE "#Results of automated BLAST query of performed on "
        . _getTime()
        . ".\n" );
print( OUTFILE
        "#Searches performed using remote_blast_client.pl, written by Paul Stothard, stothard\@ualberta.ca.\n"
);
print( OUTFILE "#The following settings were specified:\n" );
my @settingsKeys = keys(%settings);

foreach (@settingsKeys) {
    if ( defined( $settings{$_} ) ) {
        print( OUTFILE "#" . $_ . "=" . $settings{$_} . "\n" );
    }
}
print( OUTFILE "#The following attributes are separated by tabs\n" );
print( OUTFILE
        "#-------------------------------------------------------------------------------------------------------------------------------------------------\n"
);
if ( $settings{ALIGN_TYPE} eq "nucleotide" ) {
    print( OUTFILE
            "query_id\tmatch_id\tmatch_description\t\%_identity\talignment_length\tmismatches\tgap_openings\tq_start\tq_end\ts_start\ts_end\tevalue\tbit_score\n"
    );
}
elsif ( $settings{ALIGN_TYPE} eq "protein" ) {
    print( OUTFILE
            "query_id\tmatch_id\tmatch_description\t\%_identity\t%_positives\talignment_length\tmismatches\tgap_opens\tq_start\tq_end\ts_start\ts_end\tevalue\tbit_score\n"
    );
}
elsif ( $settings{ALIGN_TYPE} eq "translated" ) {
    print( OUTFILE
            "query_id\tmatch_id\tmatch_description\t\%_identity\t%_positives\tquery/sbjct_frames\talignment_length\tmismatches\tgap_opens\tq_start\tq_end\ts_start\ts_end\tevalue\tbit_score\n"
    );
}

close(OUTFILE) or die("Cannot close file : $!");

my $seqCount = 0;
local $/ = ">";
open( SEQFILE, $settings{INPUTFILE} ) or die("Cannot open file : $!");

while ( my $sequenceEntry = <SEQFILE> ) {
    $seqCount++;
    if ( $sequenceEntry eq ">" ) {
        next;
    }
    my $title    = undef;
    my $sequence = undef;
    my %query    = ();
    if ( $sequenceEntry =~ m/^([^\n\cM]+)/ ) {
        $title = $1;
    }
    else {
        $title = "No title available";
    }
    $sequence = $sequenceEntry;
    $sequence =~ s/^[^\n\cM]+//;
    $sequence =~ s/[^A-Z]//ig;
    if ( !( $sequence =~ m/[A-Z]/i ) ) {
        next;
    }
    $query{title}    = $title;
    $query{sequence} = $sequence;

    if ( defined( $settings{MIN_HSP_PROP} ) ) {
        my $queryLength = length( $query{sequence} );
        if ( $settings{ALIGN_TYPE} eq "nucleotide" ) {
            $settings{MIN_HSP_LENGTH}
                = $settings{MIN_HSP_PROP} * $queryLength;
        }
        elsif ( $settings{ALIGN_TYPE} eq "protein" ) {
            $settings{MIN_HSP_LENGTH}
                = $settings{MIN_HSP_PROP} * $queryLength;
        }
        elsif ( $settings{ALIGN_TYPE} eq "translated" ) {
            $settings{MIN_HSP_LENGTH}
                = $settings{MIN_HSP_PROP} * $queryLength / 3;
        }
        $settings{MIN_HSP_LENGTH}
            = sprintf( "%.0f", $settings{MIN_HSP_LENGTH} );
    }

    my $RIDsuccess  = 0;
    my $RIDattempts = 0;
    my $RID         = undef;
    my $RTOE        = undef;
    my $RIDresponse = undef;
    while ( ( !($RIDsuccess) ) && ( $RIDattempts < $settings{ERROR_RETRY} ) )
    {
        $RIDresponse = _getRID( \%settings, \%query );
        my $success = $RIDresponse->is_success();
        if ($success) {
            my $result = $RIDresponse->as_string();
            if ( $result
                =~ m/QBlastInfoBegin\s*RID\s=\s([^\s]+)\s*RTOE\s=\s(\d+)\s*QBlastInfoEnd/
                )
            {
                $RID  = $1;
                $RTOE = $2;
            }
            else {
                print
                    "$result\n\nError: Could not parse RID and RTOE from above response.\n";
            }
        }
        else {
            print
                "Error: Could not get response when requesting RID and RTOE for sequence "
                . $query{title} . "\n";
        }

        if ( ( defined($RID) ) && ( defined($RTOE) ) ) {
            $RIDsuccess = 1;
            print "RID received for "
                . $query{title}
                . ". The RTOE for this sequence is $RTOE seconds.\n";
        }
        else {
            $RIDattempts++;
            sleep(60);
        }
    }

    if ( !($RIDsuccess) ) {
        print "Error1: "
            . $query{title}
            . " failed at RID and RTOE requst.\n";
        open( OUTFILE, "+>>" . $settings{OUTPUTFILE} )
            or die("Cannot open file : $!");
        print(    OUTFILE "Error: "
                . $query{title}
                . " failed at RID and RTOE request.\n" );
        close(OUTFILE) or die("Cannot close file : $!");
        next;
    }

    print "Sleeping for $RTOE seconds.\n";
    sleep($RTOE);
    my $BLASTsuccess  = 0;
    my $BLASTattempts = 0;
    my $BLASTresponse = undef;
    my $BLASTresult   = undef;
    while (( !($BLASTsuccess) )
        && ( $BLASTattempts < $settings{ERROR_RETRY} ) )
    {
        $BLASTresponse = _getBLAST( \%settings, $RID );
        my $success = $BLASTresponse->is_success();
        if ($success) {
            my $result = $BLASTresponse->as_string();

            if ( $result =~ m/<td>status<\/td><td>searching<\/td>/i ) {
                print "The results are not ready for sequence "
                    . $query{title} . ".\n";
                print "Pausing before requesting again.\n";
                sleep(20);
                next;
            }
            elsif ( $result
                =~ m/QBlastInfoBegin\s*Status=WAITING\s*QBlastInfoEnd/ )
            {
                print "The results are not ready for sequence "
                    . $query{title} . ".\n";
                print "Pausing before requesting again.\n";
                sleep(20);
                next;
            }
            else {
                print "The results were received for sequence "
                    . $query{title} . ".\n";
                $BLASTresult = $result;
            }
        }
        else {
            print
                "Error: Could not get response when requesting status for sequence "
                . $query{title} . "\n";
        }

        if ( defined($BLASTresult) ) {
            $BLASTsuccess = 1;
        }
        else {
            $BLASTattempts++;
            sleep(60);
        }
    }

    if ( !($BLASTsuccess) ) {
        print "Error: "
            . $query{title}
            . " failed at BLAST results request.\n";
        open( OUTFILE, "+>>" . $settings{OUTPUTFILE} )
            or die("Cannot open file : $!");
        print(    OUTFILE "Error: "
                . $query{title}
                . " failed at BLAST results request.\n" );
        close(OUTFILE) or die("Cannot close file : $!");
        next;
    }

    my $hitFound = 0;
    my $HSPCount = 0;

    my @results;
    if ( ( defined( $settings{XML} ) ) && ( $settings{XML} =~ m/t/i ) ) {
        @results = @{ _parse_blast_xml( \%settings, $BLASTresult ) };
    }
    else {
        @results = @{ _parse_blast_table( \%settings, $BLASTresult ) };
    }

    foreach (@results) {

        my $HSP = $_;

        $HSPCount++;

        if (   ( defined( $settings{HSP_MAX} ) )
            && ( $HSPCount > $settings{HSP_MAX} ) )
        {
            next;
        }

        if (   ( defined( $settings{HSPS_PER_HIT} ) )
            && ( defined( $HSP->{hsp_number} ) )
            && ( $HSP->{hsp_number} > $settings{HSPS_PER_HIT} ) )
        {
            next;
        }

        if ( defined( $settings{MIN_HSP_LENGTH} ) ) {
            if ( $HSP->{alignment_length} < $settings{MIN_HSP_LENGTH} ) {
                print "Skipping HSP because alignment length is less than "
                    . $settings{MIN_HSP_LENGTH} . ".\n";
                next;
            }
        }

        if ( defined( $settings{MIN_SCORE} ) ) {
            if ( $HSP->{bit_score} < $settings{MIN_SCORE} ) {
                print "Skipping HSP because score is less than "
                    . $settings{MIN_SCORE} . ".\n";
                next;
            }
        }

        if ( defined( $settings{MIN_IDENTITY} ) ) {
            if ( $HSP->{identity} < $settings{MIN_IDENTITY} ) {
                print "Skipping HSP because identity is less than "
                    . $settings{MIN_IDENTITY} . ".\n";
                next;
            }
        }

        #this is to obtain a single gi number
        if ( $HSP->{match_id} =~ m/(ref|gi)\|(\d+)/ ) {
            $HSP->{uid}      = $2;
            $HSP->{match_id} = $1 . "|" . $2;
        }

        if (   ( $settings{FETCH_DESC} =~ m/t/i )
            && ( !defined( $HSP->{match_description} ) ) )
        {
            if ( !( defined( $HSP->{uid} ) ) ) {
                $HSP->{match_description} = "No identifier available";
            }
            else {
                $HSP->{match_description} = "-";
                my $ENTREZsuccess  = 0;
                my $ENTREZattempts = 0;
                my $ENTREZresponse = undef;
                my $ENTREZresult   = undef;
                while (( !($ENTREZsuccess) )
                    && ( $ENTREZattempts < $settings{ERROR_RETRY} ) )
                {
                    $ENTREZresponse = _getENTREZ( \%settings, $HSP->{uid} );
                    my $success = $ENTREZresponse->is_success();
                    if ($success) {
                        my $result = $ENTREZresponse->as_string();

                        if ( $result
                            =~ m/<Item Name="Title" Type="String">(.*?)<\/Item>/i
                            )
                        {
                            $ENTREZresult = $1;
                        }
                        else {
                            print
                                "Error: Could not parse Entrez information from response $result.\n";
                        }

                        if (   ( $settings{ACCESSION} =~ m/t/i )
                            && ( !defined( $HSP->{accession} ) ) )
                        {

#<Item Name="Caption" Type="String">AM434294</Item>
#or
#<Item Name="Extra" Type="String">gi|123673855|emb|AM434294.1|[123673855]</Item>
                            if ( $result
                                =~ m/<Item Name="Caption" Type="String">(.*?)<\/Item>/i
                                )
                            {
                                $HSP->{accession} = $1;
                            }
                            else {
                                print
                                    "Error: Could not parse accession information from response $result.\n";
                            }
                        }

                    }
                    else {
                        print
                            "Error: Could not get response when requesting Entrez information for gi "
                            . $HSP->{uid} . "\n";
                    }

                    if ( defined($ENTREZresult) ) {
                        $ENTREZsuccess = 1;
                        print "ENTREZ results received for "
                            . $HSP->{uid} . "\n";
                    }
                    else {
                        $ENTREZattempts++;
                        sleep(60);
                    }
                }

                if ( !($ENTREZsuccess) ) {
                    $HSP->{match_description}
                        = "Unable to obtain description from ENTREZ";
                }
                else {
                    $HSP->{match_description} = $ENTREZresult;
                }
            }
        }

        print "Writing HSP to file.\n";
        open( OUTFILE, "+>>" . $settings{OUTPUTFILE} )
            or die("Cannot open file : $!");

        $HSP->{id_to_return} = $HSP->{match_id};

        if (   ( $settings{ACCESSION} =~ m/t/i )
            && ( defined( $HSP->{accession} ) ) )
        {
            $HSP->{id_to_return} = $HSP->{accession};
        }

        if (   ( $settings{HSP_LABEL} =~ m/t/i )
            && ( defined( $HSP->{hit_number} ) )
            && ( defined( $HSP->{hsp_number} ) ) )
        {
            $HSP->{id_to_return}
                = $HSP->{id_to_return} . ";hit_number=$HSP->{hit_number}";
        }

        if ( !defined( $HSP->{match_description} ) ) {
            $HSP->{match_description} = "-";
        }

        if ( $settings{ALIGN_TYPE} eq "nucleotide" ) {
            print( OUTFILE
                    "$HSP->{query_id}\t$HSP->{id_to_return}\t$HSP->{match_description}\t$HSP->{identity}\t$HSP->{alignment_length}\t$HSP->{mismatches}\t$HSP->{gap_opens}\t$HSP->{q_start}\t$HSP->{q_end}\t$HSP->{s_start}\t$HSP->{s_end}\t$HSP->{evalue}\t$HSP->{bit_score}\n"
            );
        }
        elsif ( $settings{ALIGN_TYPE} eq "protein" ) {
            print( OUTFILE
                    "$HSP->{query_id}\t$HSP->{id_to_return}\t$HSP->{match_description}\t$HSP->{identity}\t$HSP->{positives}\t$HSP->{alignment_length}\t$HSP->{mismatches}\t$HSP->{gap_opens}\t$HSP->{q_start}\t$HSP->{q_end}\t$HSP->{s_start}\t$HSP->{s_end}\t$HSP->{evalue}\t$HSP->{bit_score}\n"
            );
        }
        elsif ( $settings{ALIGN_TYPE} eq "translated" ) {
            print( OUTFILE
                    "$HSP->{query_id}\t$HSP->{id_to_return}\t$HSP->{match_description}\t$HSP->{identity}\t$HSP->{positives}\t$HSP->{query_sbjct_frames}\t$HSP->{alignment_length}\t$HSP->{mismatches}\t$HSP->{gap_opens}\t$HSP->{q_start}\t$HSP->{q_end}\t$HSP->{s_start}\t$HSP->{s_end}\t$HSP->{evalue}\t$HSP->{bit_score}\n"
            );
        }

        close(OUTFILE) or die("Cannot close file : $!");
        $hitFound = 1;

    }
    if ( !($hitFound) ) {
        open( OUTFILE, "+>>" . $settings{OUTPUTFILE} )
            or die("Cannot open file : $!");
        print( OUTFILE $query{title} . "\t"
                . "no acceptable hits returned\n" );
        close(OUTFILE) or die("Cannot close file : $!");
    }

}
close(SEQFILE) or die("Cannot close file : $!");
print "Open " . $settings{OUTPUTFILE} . " to view the BLAST results.\n";

sub _getRID {
    my $settings = shift;
    my $query    = shift;
    my $fasta    = ">" . $query->{title} . "\n" . $query->{sequence};

    my $RIDresponse = undef;

    #-filter - the sequence filter to use on the query. Default is
    #'none'. Acceptable values are 'L' for low complexity, and 'R' for
    #human repeats, or 'LR' for both.

    #This script uses the 'm' filter option with all filters.
    #see http://www.ncbi.nlm.nih.gov/blast/blastcgihelp.shtml#filter

    #For more information about the BLAST URL API, see:
    #http://www.ncbi.nlm.nih.gov/blast/Doc/urlapi.html

    #2007-06-21; Added COMPOSITION_BASED_STATISTICS parameter

    if (   ( !( defined( $settings->{FILTER} ) ) )
        || ( $settings->{FILTER} eq 'F' ) )
    {
        $RIDresponse = $settings->{BROWSER}->request(
            POST(
                $settings->{BLAST_URL},
                [   DATABASE        => $settings->{DATABASE},
                    HITLIST_SIZE    => $settings->{HITLIST_SIZE},
                    ENTREZ_QUERY    => $settings->{ENTREZ_QUERY},
                    EXPECT          => $settings->{EXPECT},
                    WORD_SIZE       => $settings->{WORD_SIZE},
                    GENETIC_CODE    => $settings->{QUERY_GENETIC_CODE},
                    DB_GENETIC_CODE => $settings->{DATABASE_GENETIC_CODE},
                    PROGRAM         => $settings->{PROGRAM},
                    MEGABLAST       => $settings->{MEGABLAST},
                    COMPOSITION_BASED_STATISTICS =>
                        $settings->{COMPOSITION_BASED_STATISTICS},
                    QUERY => $fasta,
                    CMD   => "Put"
                ]
            )
        );
    }
    elsif ( $settings->{FILTER} eq 'L' ) {
        $RIDresponse = $settings->{BROWSER}->request(
            POST(
                $settings->{BLAST_URL},
                [   DATABASE        => $settings->{DATABASE},
                    HITLIST_SIZE    => $settings->{HITLIST_SIZE},
                    ENTREZ_QUERY    => $settings->{ENTREZ_QUERY},
                    EXPECT          => $settings->{EXPECT},
                    WORD_SIZE       => $settings->{WORD_SIZE},
                    FILTER          => $settings->{FILTER},
                    FILTER          => "m",
                    GENETIC_CODE    => $settings->{QUERY_GENETIC_CODE},
                    DB_GENETIC_CODE => $settings->{DATABASE_GENETIC_CODE},
                    PROGRAM         => $settings->{PROGRAM},
                    MEGABLAST       => $settings->{MEGABLAST},
                    COMPOSITION_BASED_STATISTICS =>
                        $settings->{COMPOSITION_BASED_STATISTICS},
                    QUERY => $fasta,
                    CMD   => "Put"
                ]
            )
        );
    }
    elsif ( $settings->{FILTER} eq 'R' ) {
        $RIDresponse = $settings->{BROWSER}->request(
            POST(
                $settings->{BLAST_URL},
                [   DATABASE        => $settings->{DATABASE},
                    HITLIST_SIZE    => $settings->{HITLIST_SIZE},
                    ENTREZ_QUERY    => $settings->{ENTREZ_QUERY},
                    EXPECT          => $settings->{EXPECT},
                    WORD_SIZE       => $settings->{WORD_SIZE},
                    FILTER          => $settings->{FILTER},
                    FILTER          => "m",
                    GENETIC_CODE    => $settings->{QUERY_GENETIC_CODE},
                    DB_GENETIC_CODE => $settings->{DATABASE_GENETIC_CODE},
                    PROGRAM         => $settings->{PROGRAM},
                    MEGABLAST       => $settings->{MEGABLAST},
                    COMPOSITION_BASED_STATISTICS =>
                        $settings->{COMPOSITION_BASED_STATISTICS},
                    QUERY => $fasta,
                    CMD   => "Put"
                ]
            )
        );
    }
    elsif ( $settings->{FILTER} =~ m/(RL)|(LR)/i ) {
        $RIDresponse = $settings->{BROWSER}->request(
            POST(
                $settings->{BLAST_URL},
                [   DATABASE        => $settings->{DATABASE},
                    HITLIST_SIZE    => $settings->{HITLIST_SIZE},
                    ENTREZ_QUERY    => $settings->{ENTREZ_QUERY},
                    EXPECT          => $settings->{EXPECT},
                    WORD_SIZE       => $settings->{WORD_SIZE},
                    FILTER          => "R",
                    FILTER          => "L",
                    FILTER          => "m",
                    GENETIC_CODE    => $settings->{QUERY_GENETIC_CODE},
                    DB_GENETIC_CODE => $settings->{DATABASE_GENETIC_CODE},
                    PROGRAM         => $settings->{PROGRAM},
                    MEGABLAST       => $settings->{MEGABLAST},
                    COMPOSITION_BASED_STATISTICS =>
                        $settings->{COMPOSITION_BASED_STATISTICS},
                    QUERY => $fasta,
                    CMD   => "Put"
                ]
            )
        );
    }
    else {
        die("Do not understand the specified -filter value: $settings->{FILTER}.\n"
        );
    }

    return $RIDresponse;
}

sub _getBLAST {
    my $settings = shift;
    my $RID      = shift;

    my $format_type;
    my $alignment_view;
    if ( ( defined( $settings->{XML} ) ) && ( $settings->{XML} =~ m/t/i ) ) {
        $format_type = 'XML';

        #This could probably be changed from 'Tabular' to 'Pairwise'
        $alignment_view = 'Tabular';
    }
    else {
        $format_type    = 'Text';
        $alignment_view = 'Tabular';
    }

    my $BLASTresponse = $settings->{BROWSER}->request(
        POST(
            $settings->{BLAST_URL},
            [   RID            => $RID,
                FORMAT_TYPE    => $format_type,
                ALIGNMENT_VIEW => $alignment_view,
                ALIGNMENTS     => $settings->{HITLIST_SIZE},
                DESCRIPTIONS   => $settings->{HITLIST_SIZE},
                CMD            => "Get"
            ]
        )
    );
    return $BLASTresponse;
}

sub _getENTREZ {
    my $settings       = shift;
    my $uid            = shift;
    my $ENTREZresponse = $settings->{BROWSER}->request(
        GET(      $settings->{ENTREZ_URL} . "db="
                . $settings->{ENTREZ_DB}
                . "&id=$uid&tool=remote_blast_client&retmode=xml"
        )
    );
    return $ENTREZresponse;
}

sub _getTime {
    my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst )
        = localtime(time);
    $year += 1900;

    my @days = (
        'Sunday',   'Monday', 'Tuesday', 'Wednesday',
        'Thursday', 'Friday', 'Saturday'
    );
    my @months = (
        'January',   'February', 'March',    'April',
        'May',       'June',     'July',     'August',
        'September', 'October',  'November', 'December'
    );
    my $time
        = $days[$wday] . " "
        . $months[$mon] . " "
        . sprintf( "%02d", $mday ) . " "
        . sprintf( "%02d", $hour ) . ":"
        . sprintf( "%02d", $min ) . ":"
        . sprintf( "%02d", $sec ) . " "
        . sprintf( "%04d", $year );
    return $time;
}

sub _setDefaults {
    my $blastType = shift;
    my $settings  = shift;

    #1 - Nucleotide-nucleotide BLAST (blastn)
    if ( ( $blastType =~ /^blastn$/i ) || ( $blastType eq "1" ) ) {
        $settings->{PROGRAM}                      = "blastn";
        $settings->{WORD_SIZE}                    = "11";
        $settings->{INPUTTYPE}                    = "DNA";
        $settings->{ENTREZ_DB}                    = "nucleotide";
        $settings->{ALIGN_TYPE}                   = "nucleotide";
        $settings->{COMPOSITION_BASED_STATISTICS} = '0';
    }

    #2 - Protein-protein BLAST (blastp)
    elsif ( ( $blastType =~ /^blastp$/i ) || ( $blastType eq "2" ) ) {
        $settings->{PROGRAM}                      = "blastp";
        $settings->{WORD_SIZE}                    = "3";
        $settings->{INPUTTYPE}                    = "protein";
        $settings->{ENTREZ_DB}                    = "protein";
        $settings->{ALIGN_TYPE}                   = "protein";
        $settings->{COMPOSITION_BASED_STATISTICS} = '1';
    }

    #3 - Translated query vs protein database (blastx)
    elsif ( ( $blastType =~ /^blastx$/i ) || ( $blastType eq "3" ) ) {
        $settings->{PROGRAM}                      = "blastx";
        $settings->{WORD_SIZE}                    = "3";
        $settings->{INPUTTYPE}                    = "DNA";
        $settings->{ENTREZ_DB}                    = "protein";
        $settings->{ALIGN_TYPE}                   = "translated";
        $settings->{COMPOSITION_BASED_STATISTICS} = '1';
    }

    #4 - Protein query vs translated database (tblastn)
    elsif ( ( $blastType =~ /^tblastn$/i ) || ( $blastType eq "4" ) ) {
        $settings->{PROGRAM}                      = "tblastn";
        $settings->{WORD_SIZE}                    = "3";
        $settings->{INPUTTYPE}                    = "protein";
        $settings->{ENTREZ_DB}                    = "nucleotide";
        $settings->{ALIGN_TYPE}                   = "translated";
        $settings->{COMPOSITION_BASED_STATISTICS} = '1';
    }

    #5 - Translated query vs. translated database (tblastx)
    elsif ( ( $blastType =~ /^tblastx$/i ) || ( $blastType eq "5" ) ) {
        $settings->{PROGRAM}                      = "tblastx";
        $settings->{WORD_SIZE}                    = "3";
        $settings->{INPUTTYPE}                    = "DNA";
        $settings->{ENTREZ_DB}                    = "nucleotide";
        $settings->{ALIGN_TYPE}                   = "translated";
        $settings->{COMPOSITION_BASED_STATISTICS} = '0';
    }

    #6 - Nucleotide-nucleotide BLAST (megablast).
    elsif ( ( $blastType =~ /^megablast$/i ) || ( $blastType eq "6" ) ) {
        $settings->{PROGRAM}                      = "blastn";
        $settings->{WORD_SIZE}                    = "28";
        $settings->{INPUTTYPE}                    = "DNA";
        $settings->{ENTREZ_DB}                    = "nucleotide";
        $settings->{ALIGN_TYPE}                   = "nucleotide";
        $settings->{COMPOSITION_BASED_STATISTICS} = '0';
        $settings->{MEGABLAST}                    = 'true';
    }

    else {
        die("BLAST type $blastType is not recognized.");
    }
}

sub _get_integer {
    my $value = shift;
    my $int   = undef;
    if ( ( defined($value) ) && ( $value =~ m/(\-*\d+)/ ) ) {
        $int = $1;
    }
    return $int;
}

sub _get_real {
    my $value = shift;
    my $real  = undef;
    if ( ( defined($value) ) && ( $value =~ m/(\S+)/ ) ) {
        $real = $1;
    }
    return $real;
}

sub _parse_blast_table {
    my $settings = shift;
    my $table    = shift;

    my $searchPattern;
    if ( $settings->{ALIGN_TYPE} eq "nucleotide" ) {
        $searchPattern
            = '^([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t\s]*$';
    }
    elsif ( $settings->{ALIGN_TYPE} eq "protein" ) {
        $searchPattern
            = '^([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t\s]*$';
    }
    elsif ( $settings->{ALIGN_TYPE} eq "translated" ) {
        $searchPattern
            = '^([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t]+([^\t]+)[\t\s]*$';
    }

    my @results = split( /\n/, $table );
    my @HSPs = ();

    foreach (@results) {

        if ( $_ =~ m/^\#\sFields:/ ) {
            next;
        }

        if ( !( $_ =~ m/$searchPattern/ ) ) {
            next;
        }

        my %HSP = (
            query_id           => undef,
            match_id           => undef,
            match_description  => undef,
            identity           => undef,
            positives          => undef,
            query_sbjct_frames => undef,
            alignment_length   => undef,
            mismatches         => undef,
            gap_opens          => undef,
            q_start            => undef,
            q_end              => undef,
            s_start            => undef,
            s_end              => undef,
            evalue             => undef,
            bit_score          => undef,
            uid                => undef,
            accession          => undef,
            id_to_return       => undef,
            hit_number         => undef,
            hsp_number         => undef
        );

        if ( $settings->{ALIGN_TYPE} eq "nucleotide" ) {
            $HSP{query_id}           = $1;
            $HSP{match_id}           = $2;
            $HSP{match_description}  = undef;
            $HSP{identity}           = $5;
            # $HSP{positives}          = undef;
            # $HSP{query_sbjct_frames} = undef;
            $HSP{alignment_length}   = $6;
            $HSP{mismatches}         = $7;
            $HSP{gap_opens}          = $8;
            $HSP{q_start}            = $9;
            $HSP{q_end}              = $10;
            $HSP{s_start}            = $11;
            $HSP{s_end}              = $12;
            $HSP{evalue}             = $13;
            $HSP{bit_score}          = $14;
        }
        elsif ( $settings->{ALIGN_TYPE} eq "protein" ) {
            $HSP{query_id}           = $1;
            $HSP{match_id}           = $2;
            $HSP{match_description}  = undef;
            $HSP{identity}           = $3;
            $HSP{positives}          = $4;
            $HSP{query_sbjct_frames} = undef;
            $HSP{alignment_length}   = $5;
            $HSP{mismatches}         = $6;
            $HSP{gap_opens}          = $7;
            $HSP{q_start}            = $8;
            $HSP{q_end}              = $9;
            $HSP{s_start}            = $10;
            $HSP{s_end}              = $11;
            $HSP{evalue}             = $12;
            $HSP{bit_score}          = $13;
        }
        elsif ( $settings->{ALIGN_TYPE} eq "translated" ) {
            $HSP{query_id}           = $1;
            $HSP{match_id}           = $2;
            $HSP{match_description}  = undef;
            $HSP{identity}           = $3;
            $HSP{positives}          = $4;
            $HSP{query_sbjct_frames} = $5;
            $HSP{alignment_length}   = $6;
            $HSP{mismatches}         = $7;
            $HSP{gap_opens}          = $8;
            $HSP{q_start}            = $9;
            $HSP{q_end}              = $10;
            $HSP{s_start}            = $11;
            $HSP{s_end}              = $12;
            $HSP{evalue}             = $13;
            $HSP{bit_score}          = $14;
        }

        push( @HSPs, \%HSP );
    }
    return \@HSPs;
}

sub _parse_blast_xml {
    my $settings = shift;
    my $xml      = shift;
    my @HSPs     = ();

    #remove header text
    $xml =~ s/^[\s\S]*?(?=<\?xml version=)//;

    #check for complete document
    if ( !( $xml =~ m/<\/BlastOutput>\s*$/ ) ) {
        print "Error: Incomplete XML results returned.\n";
        return \@HSPs;
    }

    my $parser = new XML::DOM::Parser();
    my $doc    = $parser->parsestring($xml);

    my $query_id
        = $doc->getElementsByTagName('BlastOutput_query-def')->item(0)
        ->getFirstChild->getNodeValue;

    for my $hit ( $doc->getElementsByTagName('Hit') ) {
        my $match_id = $hit->getElementsByTagName('Hit_id')->item(0)
            ->getFirstChild->getNodeValue;
        my $match_description = $hit->getElementsByTagName('Hit_def')->item(0)
            ->getFirstChild->getNodeValue;
        my $accession = $hit->getElementsByTagName('Hit_accession')->item(0)
            ->getFirstChild->getNodeValue;
        my $hit_number = $hit->getElementsByTagName('Hit_num')->item(0)
            ->getFirstChild->getNodeValue;
        for my $hsp ( $hit->getElementsByTagName('Hsp') ) {
            my %HSP = (
                query_id           => undef,
                match_id           => undef,
                match_description  => undef,
                identity           => undef,
                positives          => undef,
                query_sbjct_frames => undef,
                alignment_length   => undef,
                mismatches         => undef,
                gap_opens          => undef,
                q_start            => undef,
                q_end              => undef,
                s_start            => undef,
                s_end              => undef,
                evalue             => undef,
                bit_score          => undef,
                uid                => undef,
                accession          => undef,
                id_to_return       => undef,
                hit_number         => undef,
                hsp_number         => undef
            );

            $HSP{query_id}          = $query_id;
            $HSP{match_id}          = $match_id;
            $HSP{match_description} = $match_description;

            my $query_frame = 0;
            my $hit_frame   = 0;
            if (defined(
                    $hsp->getElementsByTagName('Hsp_query-frame')->item(0)
                )
                )
            {
                $query_frame
                    = $hsp->getElementsByTagName('Hsp_query-frame')->item(0)
                    ->getFirstChild->getNodeValue;
            }
            if (defined(
                    $hsp->getElementsByTagName('Hsp_hit-frame')->item(0)
                )
                )
            {
                $hit_frame
                    = $hsp->getElementsByTagName('Hsp_hit-frame')->item(0)
                    ->getFirstChild->getNodeValue;
            }
            $HSP{query_sbjct_frames} = $query_frame . '/' . $hit_frame;

            $HSP{alignment_length}
                = $hsp->getElementsByTagName('Hsp_align-len')->item(0)
                ->getFirstChild->getNodeValue;

            #percent identity
            my $identity = $hsp->getElementsByTagName('Hsp_identity')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{identity} = 100 * $identity / $HSP{alignment_length};

            #percent positives
            my $positives
                = $hsp->getElementsByTagName('Hsp_positive')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{positives} = 100 * $positives / $HSP{alignment_length};

         #The 'Hsp_gaps' tag gives the number of gaps, whereas the 'gap_opens'
         #value in the tabular output is the number of gap openings. Need to
         #count gap openings.
            $HSP{gap_opens} = 0;

            my $Hsp_qseq = $hsp->getElementsByTagName('Hsp_qseq')->item(0)
                ->getFirstChild->getNodeValue;

            my $Hsp_hseq = $hsp->getElementsByTagName('Hsp_hseq')->item(0)
                ->getFirstChild->getNodeValue;

            while ( $Hsp_qseq =~ m/\-+/g ) {
                $HSP{gap_opens}++;
            }
            while ( $Hsp_hseq =~ m/\-+/g ) {
                $HSP{gap_opens}++;
            }

            my $gaps = 0;
            if ( defined( $hsp->getElementsByTagName('Hsp_gaps')->item(0) ) )
            {
                $gaps = $hsp->getElementsByTagName('Hsp_gaps')->item(0)
                    ->getFirstChild->getNodeValue;
            }

            $HSP{mismatches} = $HSP{alignment_length} - $identity - $gaps;

            $HSP{q_start}
                = $hsp->getElementsByTagName('Hsp_query-from')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{q_end} = $hsp->getElementsByTagName('Hsp_query-to')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{s_start}
                = $hsp->getElementsByTagName('Hsp_hit-from')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{s_end} = $hsp->getElementsByTagName('Hsp_hit-to')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{evalue} = $hsp->getElementsByTagName('Hsp_evalue')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{bit_score}
                = $hsp->getElementsByTagName('Hsp_bit-score')->item(0)
                ->getFirstChild->getNodeValue;
            $HSP{accession}  = $accession;
            $HSP{hit_number} = $hit_number;
            $HSP{hsp_number} = $hsp->getElementsByTagName('Hsp_num')->item(0)
                ->getFirstChild->getNodeValue;

            #round some values
            $HSP{bit_score} = sprintf( "%.2f", $HSP{bit_score} );
            $HSP{identity}  = sprintf( "%.2f", $HSP{identity} );
            $HSP{positives} = sprintf( "%.2f", $HSP{positives} );

       #NCBI uses different conventions to indicate which strand of a sequence
       #produced a hit. Adjust the values here to match the table output.
            if ( $settings->{PROGRAM} eq 'blastn' ) {

     #Make sure opposite strand match is indicated by subject positions rather
     #than by query positions
                if ( $HSP{q_start} > $HSP{q_end} ) {
                    ( $HSP{q_start}, $HSP{q_end} )
                        = ( $HSP{q_end}, $HSP{q_start} );
                    ( $HSP{s_start}, $HSP{s_end} )
                        = ( $HSP{s_end}, $HSP{s_start} );
                }

            }
            if (   ( $settings->{PROGRAM} eq 'blastx' )
                || ( $settings->{PROGRAM} eq 'tblastx' ) )
            {

                #Make sure q_start > q_end when Hsp_query-frame is negative
                if ( ( $query_frame < 0 ) && ( $HSP{q_start} < $HSP{q_end} ) )
                {
                    ( $HSP{q_start}, $HSP{q_end} )
                        = ( $HSP{q_end}, $HSP{q_start} );
                }

                #Make sure q_start < q_end when Hsp_query-frame is positive
                if ( ( $query_frame > 0 ) && ( $HSP{q_start} > $HSP{q_end} ) )
                {
                    ( $HSP{q_start}, $HSP{q_end} )
                        = ( $HSP{q_end}, $HSP{q_start} );
                }
            }
            if (   ( $settings->{PROGRAM} eq 'tblastn' )
                || ( $settings->{PROGRAM} eq 'tblastx' ) )
            {

                #Make sure s_start > s_end when Hsp_hit-frame is negative
                if ( ( $hit_frame < 0 ) && ( $HSP{s_start} < $HSP{s_end} ) ) {
                    ( $HSP{s_start}, $HSP{s_end} )
                        = ( $HSP{s_end}, $HSP{s_start} );
                }

                #Make sure s_start < s_end when Hsp_hit-frame is positive
                if ( ( $hit_frame > 0 ) && ( $HSP{s_start} > $HSP{s_end} ) ) {
                    ( $HSP{s_start}, $HSP{s_end} )
                        = ( $HSP{s_end}, $HSP{s_start} );
                }
            }

            push( @HSPs, \%HSP );
        }
    }
    $doc->dispose;
    return \@HSPs;

}
